#!/usr/bin/env python

import os, sys
from code import InteractiveInterpreter as Interpreter
from qt import *

class PyCute(QTextEdit):

    """
    PyCute is a Python shell for PyQt.

    Creating, displaying and controlling PyQt widgets from the Python command
    line interpreter is very hard, if not, impossible.  PyCute solves this
    problem by interfacing the Python interpreter to a PyQt widget.

    My use is interpreter driven plotting to QwtPlot instances. Why?
    
    Other popular scientific software packages like SciPy, SciLab, Octave,
    Maple, Mathematica, GnuPlot, ..., also have interpreter driven plotting.  
    It is well adapted to quick & dirty exploration. 

    PyCute takes 4 keywords arguments:

    - full   -- if full:
                    PyCute sees Python's full environment
                else:
                    PyCute sees a restricted environment
    - eof    -- if eof:
                    Python exits on Ctrl-D or Ctrl-Z (OS dependent)
                else:
                    import sys; sys.exit() will stop Python
    - gnu    -- if gnu:
                    GNU command history recall policy (LIFO)
                else:
                    context sensitive recall policy (closest commands first)
    - parent -- an optional parent PyQt Widget
    
    Of course, PyQt's debugger -- eric -- gives you similar facilities, but
    PyCute is smaller and easier to integrate in applications.

    PyCute is based on ideas and code from:
    - Python*/Tools/idle/PyShell.py (Python Software Foundation License)
    - PyQt*/eric/Shell.py (Gnu Public License)
    """
    
    def __init__(self, full=None, eof=None, gnu=1, parent=None):

        QTextEdit.__init__(self, parent)

        # interpreter capabilities
        if full:
            self.interpreter = Interpreter(sys.modules['__main__'].__dict__)
        else:
            self.interpreter = Interpreter()

        # history recall policy
        self.gnu = gnu

        # to exit the interpreter by a keyboard shortcut
        if eof:
             if os.name == 'posix':
                 self.eofKey = Qt.Key_D
             elif os.name == 'nt' or os.name == 'dos':
                 self.eofKey = Qt.Key_Z
             else:
                 raise SystemExit, "FIXME for 'os2', 'mac', 'ce' or 'riscos'"
        else:
            self.eofKey = None

        # capture all interactive input/output 
        sys.stdout   = self
        sys.stderr   = self
        sys.stdin    = self
        # last line + last incomplete lines
        self.line    = QString()
        self.lines   = []
        # the cursor position in the last line
        self.point   = 0
        # flag: the interpreter needs more input to run the last lines. 
        self.more    = 0
        # flag: readline() is being used for e.g. raw_input() and input()
        self.reading = 0
        # history
        self.history = []
        self.pointer = 0
        self.xLast   = 0
        self.yLast   = 0

        # user interface setup
        self.setTextFormat(QTextEdit.PlainText)
        self.setWrapPolicy(QTextEdit.Anywhere)
        self.setCaption('PyCute -- a Python Shell for PyQt')
        font = QFont("Fixed", 12)
        font.setFixedPitch(1)
        self.setFont(font)

        # geometry
        height = 40*QFontMetrics(font).lineSpacing()
        request = QSize(600, height)
        if parent is not None:
            request = request.boundedTo(parent.size())
        self.resize(request)

        # interpreter prompt.
        try:
            sys.ps1
        except AttributeError:
            sys.ps1 = ">>> "
        try:
            sys.ps2
        except AttributeError:
            sys.ps2 = "... "

        # interpreter banner
        self.write('The PyCute shell running Python %s on %s.\n' %
                   (sys.version, sys.platform))
        self.write('The author of PyCute is Gerard Vermeulen'
                   ' <http://gerard.vermeulen.free.fr>.\n')
        self.write('Type "copyright", "credits" or "license"'
                   ' for more information on Python.\n')
        self.write(sys.ps1)
        
    def flush(self):
        """
        Simulate stdin, stdout, and stderr.
        """
        pass

    def isatty(self):
        """
        Simulate stdin, stdout, and stderr.
        """
        return 1

    def readline(self):
        """
        Simulate stdin, stdout, and stderr.
        """
        self.reading = 1
        self.__clearLine()
        self.moveCursor(QTextEdit.MoveEnd, 0)
        while self.reading:
            qApp.processOneEvent()
        if self.line.length() == 0:
            return '\n'
        else:
            return str(self.line) 
    
    def write(self, text):
        """
        Simulate stdin, stdout, and stderr.
        """
        # The output of self.append(text) contains to many newline characters,
        # so work around QTextEdit's policy for handling newline characters.
        hack = self.text()
        hack.append(text)
        self.setText(hack)
        #self.setText(self.text().append(text)) # segmentation fault
        self.moveCursor(QTextEdit.MoveEnd, 0)
        self.yLast, self.xLast = self.getCursorPosition()

    def writelines(self, text):
        """
        Simulate stdin, stdout, and stderr.
        """
        map(self.write, text)
        print "DO WE EVER GET HERE? IF YES, OPTIMIZATION POSSIBLE"

    def fakeUser(self, lines):
        """
        Simulate a user: lines is a sequence of strings, (Python statements).
        """
        for line in lines:
            self.line = QString(line.rstrip())
            self.write(self.line)
            self.write('\n')
            self.__run()
            
    def __run(self):
        """
        Append the last line to the history list, let the interpreter execute
        the last line(s), and clean up accounting for the interpreter results:
        (1) the interpreter succeeds
        (2) the interpreter fails, finds no errors and wants more line(s)
        (3) the interpreter fails, finds errors and writes them to sys.stderr
        """
        if self.gnu:
            self.pointer = 0
        self.history.append(QString(self.line))
        self.lines.append(str(self.line))
        source = '\n'.join(self.lines)
        self.more = self.interpreter.runsource(source)
        if self.more:
            self.write(sys.ps2)
        else:
            self.write(sys.ps1)
            self.lines = []
        self.__clearLine()
        
    def __clearLine(self):
        """
        Clear input line buffer
        """
        self.line.truncate(0)
        self.point = 0
        
    def __insertText(self, text):
        """
        Insert text at the current cursor position.
        """
        y, x = self.getCursorPosition()
        self.insertAt(text, y, x)
        self.line.insert(self.point, text)
        self.point += text.length()
        self.setCursorPosition(y, x + text.length())

    def keyPressEvent(self, e):
        """
        Handle user input a key at a time.
        """
        text  = e.text()
        key   = e.key()
        ascii = e.ascii()

        if text.length() and ascii>=32 and ascii<127:
            self.__insertText(text)
            return

        if e.state() & Qt.ControlButton and key == self.eofKey:
            sys.exit()
            return
        
        if e.state() & Qt.ControlButton or e.state() & Qt.ShiftButton:
            e.ignore()
            return

        if key == Qt.Key_Backspace:
            if self.point:
                self.doKeyboardAction(QTextEdit.ActionBackspace)
                self.point -= 1
                self.line.remove(self.point, 1)
        elif key == Qt.Key_Delete:
            self.doKeyboardAction(QTextEdit.ActionDelete)
            self.line.remove(self.point, 1)
        elif key == Qt.Key_Return or key == Qt.Key_Enter:
            self.write('\n')
            if self.reading:
                self.reading = 0
            else:
                self.__run()
        elif key == Qt.Key_Tab:
            self.__insertText(text)
        elif key == Qt.Key_Left:
            if self.point:
                self.moveCursor(QTextEdit.MoveBackward,0)
                self.point -= 1
        elif key == Qt.Key_Right:
            if self.point < self.line.length():
                self.moveCursor(QTextEdit.MoveForward,0)
                self.point += 1
        elif key == Qt.Key_Home:
            self.setCursorPosition(self.yLast, self.xLast)
            self.point = 0
        elif key == Qt.Key_End:
            self.moveCursor(QTextEdit.MoveLineEnd,0)
            self.point = self.line.length()
        elif key == Qt.Key_Up:
            if len(self.history):
                if self.pointer == 0:
                    self.pointer = len(self.history)
                self.pointer -= 1
                self.__recall()
        elif key == Qt.Key_Down:
            if len(self.history):
                self.pointer += 1
                if self.pointer == len(self.history):
                    self.pointer = 0
                self.__recall()
        else:
            e.ignore()

    def __recall(self):
        """
        Display the current item from the command history.
        """
        self.setCursorPosition(self.yLast, self.xLast)
        self.setSelection(self.yLast, self.xLast,
                          self.yLast, self.paragraphLength(self.yLast))
        self.removeSelectedText()
        self.__clearLine()
        self.__insertText(self.history[self.pointer])

        
    def focusNextPrevChild(self, next):
        """
        Suppress tabbing to the next window in multi-line commands. 
        """
        if next and self.more:
            return 0
        return QTextEdit.focusNextPrevChild(self, next)

    def mousePressEvent(self, e):
        """
        Keep the cursor after the last prompt.
        """
        if e.button() == Qt.LeftButton:
            self.moveCursor(QTextEdit.MoveEnd, 0)
        return

    def contentsContextMenuEvent(self,ev):
        """
        Suppress the right button context menu.
        """
        return


# Admire
if __name__ == '__main__':
    
    # The next statements assure importation of qwt from the
    # local source tree. This is useful for debugging and testing.
    import glob
    globs = glob.glob('../build/lib.*/qwt')
    if globs:                           # found PyQwt modules made by distutils
        sys.path.insert(0, os.path.split(globs[0])[0])

    # import a lot of useful stuff
    from Numeric import *
    from qwt import *
    from plot import *
    
    ___a___ = QApplication(sys.argv)
    
    # full=1   -- make the full Python environment visible to the PyCute shell
    # eof=1    -- make the keyboard shortcuts Ctrl-d or Ctrl-z exit Python 
    # gnu=None -- choose context sensitive history 
    ___w___ = PyCute(full=1, eof=1, gnu=None)
    ___a___.setMainWidget(___w___)
    ___w___.show()

    # PyCute can do without users!
    ___w___.fakeUser([
        'x = arange(-2*pi, 2*pi, 0.01)',
        'y1 = sin(x)',
        'y2 = cos(x)',
        'c1 = Curve(x, y1, red, "sin(x)")',
        'c2 = Curve(x, y2, blue, "cos(x)")',
        'p = Plot(title="Plot generated with PyCute", curves=[c1, c2])',
        ])

    # User, it is your turn!
    ___a___.exec_loop()
